#	mailcoil - Effortless, featureful SMTP
#	Copyright (C) 2011-2025 Johannes Bauer
#
#	This file is part of mailcoil.
#
#	mailcoil is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	mailcoil is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with mailcoil; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import datetime
import enum
import urllib.parse
import time
import smtplib
import imaplib
from .Exceptions import MaildropFailedException
from .SerializedEmail import SerializedEmail

class MailDropoff():
	class Scheme(enum.Enum):
		LMTP = "lmtp"
		LMTP_STARTTLS = "lmtp+starttls"
		SMTP = "smtp"
		SMTPS = "smtps"
		SMTP_STARTTLS = "smtp+startls"
		IMAP = "imap"
		IMAPS = "imaps"
		FILE = "file"

	_SCHEME_BY_NAME = { scheme.value: scheme for scheme in Scheme }

	def __init__(self, scheme: Scheme, host: str, port: int | None = None, username: str | None = None, password: str | None = None, path: str | None = None):
		self._scheme = scheme
		self._host = host
		if port is None:
			self._port = {
				self.Scheme.LMTP:			24,
				self.Scheme.LMTP_STARTTLS:	24,
				self.Scheme.SMTP:			25,
				self.Scheme.SMTPS:			465,
				self.Scheme.SMTP_STARTTLS:	25,
				self.Scheme.IMAP:			143,
				self.Scheme.IMAPS:			993,
				self.Scheme.FILE:			None,
			}[self._scheme]
		else:
			self._port = port
		self._username = username
		self._password = password
		self._path = path or ""
		if self._scheme != self.Scheme.FILE:
			self._path = self._path.lstrip("/")
		if self._path == "":
			self._path = None
		if (self._scheme in [ self.Scheme.IMAP, self.Scheme.IMAPS ]) and (self._path is None):
			raise ValueError("For storing an email on IMAP, you need to specify a mailbox folder as a path.")

	@classmethod
	def parse_uri(cls, uri: str):
		parsed = urllib.parse.urlparse(uri)
		if parsed.scheme not in cls._SCHEME_BY_NAME:
			raise ValueError(f"\"{parsed.scheme}\" is not a valid URI scheme, supported are: {', '.join(sorted(cls._SCHEME_BY_NAME))}")
		scheme = cls._SCHEME_BY_NAME[parsed.scheme]
		if ":" in parsed.netloc:
			(host, port) = parsed.netloc.split(":", maxsplit = 1)
			port = int(port)
		else:
			(host, port) = (parsed.netloc, None)

		return cls(scheme = scheme, host = host, port = port, path = parsed.path)

	@property
	def username(self):
		return self._username

	@username.setter
	def username(self, value: str):
		self._username = value

	@property
	def password(self):
		return self._password

	@password.setter
	def password(self, value: str):
		self._password = value

	def _postall_smtp(self, serialized_mails: list["SerializedEmail"]):
		conn_class = smtplib.SMTP_SSL if (self._scheme == self.Scheme.SMTPS) else smtplib.SMTP
		with conn_class(self._host, self._port) as conn:
			try:
				if self._scheme == self.Scheme.SMTP_STARTTLS:
					conn.starttls()
				if (self._username is not None) and (self._password is not None):
					conn.login(self._username, self._password)

				for serialized_mail in serialized_mails:
					conn.send_message(serialized_mail.content, to_addrs = serialized_mail.recipients)
			finally:
				conn.quit()

	def _postall_imap(self, serialized_mails: list["SerializedEmail"]):
		conn_class = imaplib.IMAP4_SSL if (self._scheme == self.Scheme.IMAPS) else imaplib.IMAP4
		with conn_class(self._host, self._port) as conn:
			try:
				if (self._username is not None) and (self._password is not None):
					try:
						conn.login(self._username, self._password)
					except imaplib.IMAP4.error as e:
						raise MaildropFailedException(f"Login to {self._scheme.name} server {self._host}:{self._port} failed: {str(e)}") from e
				else:
					raise MaildropFailedException("IMAP delivery requires authentication.")

				(status, imap_rsp) = conn.select(mailbox = self._path)
				if status != "OK":
					raise MaildropFailedException(f"No such IMAP mailbox \"{self._path}\" on {self._host}:{self._port}: {str(imap_rsp)}")

				for serialized_mail in serialized_mails:
					imap_date_time = imaplib.Time2Internaldate(time.time())
					(status, imap_rsp) = conn.append(mailbox = self._path, flags = None, date_time = imap_date_time, message = bytes(serialized_mail.content))
					if status != "OK":
						raise MaildropFailedException(f"Unable to append message to \"{self._path}\" on {self._host}:{self._port}: {str(imap_rsp)}")
			finally:
				conn.logout()

	def _postall_file(self, serialized_mails: list["SerializedEmail"]):
		post_date = datetime.datetime.now().strftime("%a %b %d %H:%M:%S %Y")
		with open(self._path, "a") as f:
			for serialized_mail in serialized_mails:
				print(f"From - {post_date}", file = f)
				print("X-Mozilla-Status: 0000", file = f)
				f.write(bytes(serialized_mail.content).decode("utf-8"))
				print(file = f)

	def postall(self, mails: list["Email"]):
		serialized_mails = [ mail if isinstance(mail, SerializedEmail) else mail.serialize() for mail in mails ]
		if self._scheme in [ self.Scheme.LMTP, self.Scheme.LMTP_STARTTLS, self.Scheme.SMTP, self.Scheme.SMTPS, self.Scheme.SMTP_STARTTLS ]:
			return self._postall_smtp(serialized_mails)
		elif self._scheme in [ self.Scheme.IMAP, self.Scheme.IMAPS ]:
			return self._postall_imap(serialized_mails)
		elif self._scheme == self.Scheme.FILE:
			return self._postall_file(serialized_mails)
		else:
			raise NotImplementedError(self._scheme)

	def post(self, mail: "Email"):
		return self.postall([ mail ])

	def __str__(self):
		return f"MailDropoff<{self._scheme.value}: {self._host}:{self._port}>"



#	mailcoil - Effortless, featureful SMTP
#	Copyright (C) 2011-2025 Johannes Bauer
#
#	This file is part of mailcoil.
#
#	mailcoil is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	mailcoil is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with mailcoil; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
"""Package for easily creating and sending mails via SMTP or SMTPS.

For usage of mailcoil you first need to define a dropoff at which mails should
be posted. This is typically your MTA or smart relay:

	dropoff = mailcoil.MailDropoff(mailcoil.MailDropoff.Scheme.SMTPS, "smtp.my-server.com")

You then can create an email:

	mail = mailcoil.Email(from_address = "Johannes Bauer <johannes.bauer@gmx.de>", subject = "What's up?").to("someone@gmx.de")

CC and BCC are also supported using the same syntax as the `.to()` method:

	mail.cc("cc1@gmx.de", "cc2@gmx.de").bcc("bcc@gmx.de")

Emails can use text/plain, text/html, or both bodies:

	mail.text = "This is the text content!"
	mail.html = "<html><body><h1>Hey!</h1><p>This is the HTML content!</body></html>"

To attach files to this mail, simply you can use the `attach` method:

	mail.attach("foo.jpg")

By default, the Content-Disposition of attachments is "attachment". However,
you can also create "inline" attachments that you can then reuse in your HTML
portion, e.g.:

	src_cid = mail.attach("foo.jpg", inline = True)
	mail.html = f"<html><body><h1>Hey!</h1><p>This is the HTML content!<img src='${src_cid}'></body></html>"

By default, the Content-ID is auto-generated, which means that it depends on
the number of attachments. If you want fixed CIDs, you can also do that. Note
that You are then responsible for not creating collisions in the CIDs:

	src_cid = mail.attach("foo.jpg", inline = True, cid = "foobar")
	mail.html = f"<html><body><h1>Hey!</h1><p>This is the HTML content!<img src='cid:foobar'></body></html>"

If you wish to sign or encrypt mail using CMS (commonly known as S/MIME), you
need to set the security property of your email and configure it properly. For
example, for just signing, this works:

	mail.security = mailcoil.CMS().sign(signer_keyfile = "my_key.pem", signer_certfile = "my_cert.pem", ca_certfile = "my_ca.pem")

If you additionally want to encrypt for three different targets, you can also
do:

	mail.security.encrypt("target1.pem", "target2.pem", "target3.pem")

Similarly, you may also only encrypt without signing:

	mail.security = mailcoil.CMS().encrypt("target.pem")

When your mail is finally set up, you can drop it off at the previously defined
dropoff:

	dropoff.post(mail)

A complete example is therefore:

	import mailcoil
	dropoff = mailcoil.MailDropoff(mailcoil.MailDropoff.Scheme.SMTPS, "smtp.my-server.com")
	mail = mailcoil.Email(from_address = "Johannes Bauer <johannes.bauer@gmx.de>", subject = "What's up?").to("someone@gmx.de")
	mail.cc("cc1@gmx.de", "cc2@gmx.de").bcc("bcc@gmx.de")
	mail.text = "This is the text content!"

	# Attach file and use it in HTML portion
	src_cid = mail.attach("foo.jpg", inline = True)
	mail.html = f"<html><body><h1>Hey!</h1><p>This is the HTML content!<img src='${src_cid}'></body></html>"

	# Sign and encrypt
	mail.security = mailcoil.CMS().sign(signer_keyfile = "my_key.pem", signer_certfile = "my_cert.pem", ca_certfile = "my_ca.pem").encrypt("target1.pem", "target2.pem", "target3.pem")

	# Send mail through the dropoff
	dropoff.post(mail)

Another, shorter example, that showcases local debugging capabilities and will
write a local file /tmp/mailbox.txt that is compatible with Thunderbird:

	import mailcoil
	dropoff = mailcoil.MailDropoff.parse_uri("file:///tmp/mailbox.txt")
	mail = mailcoil.Email(from_address = "Johannes Bauer <johannes.bauer@gmx.de>", subject = "What's up?").to("someone@gmx.de")
	mail.text = "This is a test email."
	dropoff.post(mail)
"""

from .MailDropoff import MailDropoff
from .Email import Email, MailAddress
from .CMS import CMS
from .Exceptions import MailCoilException

VERSION = "0.0.7"



#	mailcoil - Effortless, featureful SMTP
#	Copyright (C) 2011-2025 Johannes Bauer
#
#	This file is part of mailcoil.
#
#	mailcoil is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	mailcoil is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with mailcoil; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import os
import email
import time
import textwrap
import dataclasses
import mimetypes
import email.policy
import email.parser
import mailcoil
from .Exceptions import NoRecipientException, NoBodyException
from .SerializedEmail import SerializedEmail

@dataclasses.dataclass
class MailAddress():
	mail: str
	name: str | None = None

	def encode(self):
		if self.name is not None:
			return email.utils.formataddr((self.name, self.mail))
		else:
			return self.mail

	@classmethod
	def parse(cls, addr: "MailAddress | str"):
		if isinstance(addr, cls):
			return addr
		elif isinstance(addr, dict):
			return MailAddress(mail = addr.get("mail"), name = addr.get("name"))
		else:
			(name, mail) = email.utils.parseaddr(addr)
			if name == "":
				return cls(mail = mail)
			else:
				((name_bindata, encoding), ) = email.header.decode_header(name)
				if encoding is None:
					return cls(mail = mail, name = name)
				else:
					return cls(mail = mail, name = name_bindata.decode(encoding))

	@classmethod
	def parsemany(cls, addrs: str):
		return [ cls.parse(addr) for addr in addrs.split(",") ]

@dataclasses.dataclass(slots = True)
class Attachment():
	data: bytes
	maintype: str
	subtype: str
	filename: str
	inline: bool
	content_id: str

class Email():
	def __init__(self, from_address: MailAddress | str, subject: str | None = None, text: str | None = None, wrap_text: bool = False, html: str | None = None, security: "SMIME | None" = None):
		self._from = MailAddress.parse(from_address)
		self._to = [ ]
		self._cc = [ ]
		self._bcc = [ ]
		self._subject = subject
		self._text = text
		self._wrap_text = wrap_text
		self._html = html
		self._security = security
		self._datetime = time.time()
		self._message_id = f"<{os.urandom(16).hex()}@mailcoil>"
		self._attachments = [ ]
		self._user_agent = None

	@property
	def recipient_count(self):
		return len(self._to) + len(self._cc) + len(self._bcc)

	def to(self, *mail_addresses: tuple[MailAddress | str]):
		self._to += [ MailAddress.parse(addr) for addr in mail_addresses ]
		return self

	def cc(self, *mail_addresses: tuple[MailAddress | str]):
		self._cc += [ MailAddress.parse(addr) for addr in mail_addresses ]
		return self

	def bcc(self, *mail_addresses: tuple[MailAddress | str]):
		self._bcc += [ MailAddress.parse(addr) for addr in mail_addresses ]
		return self

	@property
	def user_agent(self):
		return self._user_agent

	@user_agent.setter
	def user_agent(self, value: str):
		self._user_agent = value

	@property
	def subject(self):
		return self._subject

	@subject.setter
	def subject(self, value: str):
		self._subject = value

	@property
	def text(self):
		return self._text

	@text.setter
	def text(self, value: str):
		self._text = value

	@property
	def html(self):
		return self._html

	@html.setter
	def html(self, value: str):
		self._html = value

	@property
	def wrapped_text(self):
		if self._wrap_text:
			return self._wrap_paragraphs(self.text)
		else:
			return self.text

	@property
	def security(self):
		return self._security

	@security.setter
	def security(self, value: str):
		self._security = value

	def _mimetype(self, filename: str, override: str | None):
		if override is None:
			(mimetype, _) = mimetypes.guess_type(filename)
			if mimetype is None:
				return "application/octet-stream"
			else:
				return mimetype
		else:
			return override

	def attach_data(self, data: bytes, filename: str, mimetype: str | None = None, inline: bool = False, cid: str | None = None):
		(maintype, subtype) = self._mimetype(filename, mimetype).split("/")
		content_id = f"cid{len(self._attachments)}" if (cid is None) else cid
		attachment = Attachment(data = data, maintype = maintype, subtype = subtype, filename = filename, inline = inline, content_id = content_id)
		self._attachments.append(attachment)
		return f"cid:{attachment.content_id}"

	def attach(self, filename: str, mimetype: str | None = None, shown_filename: str | None = None, inline: bool = False, cid: str | None = None):
		with open(filename, "rb") as f:
			data = f.read()
		if shown_filename is None:
			shown_filename = os.path.basename(filename)
		return self.attach_data(data, filename = shown_filename, mimetype = mimetype, inline = inline, cid = cid)

	@staticmethod
	def _wrap_paragraphs(text: str) -> str:
		wrapped = [ ]
		for paragraph in text.split("\n"):
			parwrapped = textwrap.wrap(paragraph, width = 72)
			if len(parwrapped) == 0:
				parwrapped = [ "" ]
			wrapped += parwrapped
		return "\n".join(wrapped)

	def _layer_text_content(self):
		if self.recipient_count == 0:
			raise NoRecipientException("Mail has no To, CC, or BCC set. Unable to serialize.")

		if (self.text is None) and (self.html is None):
			raise NoBodyException("Mail has no text or HTML content.")

		msg = email.message.EmailMessage()

		if (self.text is not None) and (self.html is not None):
			# text and HTML as multipart/alternative
			msg.set_content(self.wrapped_text, subtype = "plain")
			msg.add_alternative(self.html, subtype = "html")
		elif self.html is None:
			msg.set_content(self.wrapped_text, subtype = "plain")
		else:
			# HTML only
			msg.set_content(self.html, subtype = "html")
		return msg

	def _layer_attachments(self, msg: "EmailMessage"):
		for attachment in self._attachments:
			if not attachment.inline:
				msg.add_attachment(attachment.data, maintype = attachment.maintype, subtype = attachment.subtype, filename = attachment.filename)
			else:
				msg.add_attachment(attachment.data, maintype = attachment.maintype, subtype = attachment.subtype, filename = attachment.filename, disposition = "inline", cid = attachment.content_id)
		return msg

	def _layer_security(self, prev_layer: "EmailMessage"):
		if self._security is None:
			msg = prev_layer
		else:
			msg = self._security.process(prev_layer)
		return msg

	def serialize(self):
		msg = self._layer_text_content()
		msg = self._layer_attachments(msg)
		msg = self._layer_security(msg)

		if self._subject is not None:
			msg["Subject"] = self._subject
		msg["Message-ID"] = self._message_id
		msg["Date"] = email.utils.formatdate(self._datetime, localtime = True)
		if self.user_agent is None:
			msg["User-Agent"] = f"mailcoil v{mailcoil.VERSION}"
		else:
			msg["User-Agent"] = f"{self.user_agent} via mailcoil v{mailcoil.VERSION}"
		msg["From"] = self._from.encode()
		if len(self._to) > 0:
			msg["To"] = ", ".join([address.encode() for address in self._to ])
		if len(self._cc) > 0:
			msg["CC"] = ", ".join([address.encode() for address in self._cc ])
		if len(self._bcc) > 0:
			msg["BCC"] = ", ".join([address.encode() for address in self._bcc ])
		return SerializedEmail(content = msg, recipients = [ addr.mail for addr in (self._to + self._cc + self._bcc) ])

	@classmethod
	def serialize_from_email_message(cls, parsed_msg: "email.message.EmailMessage"):
		def _extract_addrs(header_value):
			if header_value is None:
				return [ ]
			return [ email_addr for (email_name, email_addr) in email.utils.getaddresses([ header_value ]) ]
		recipients = _extract_addrs(parsed_msg["To"]) + _extract_addrs(parsed_msg["CC"]) + _extract_addrs(parsed_msg["BCC"])
		return SerializedEmail(content = parsed_msg, recipients = recipients)

	@classmethod
	def serialize_from_bytes(cls, email_msg: bytes):
		parsed_msg = email.parser.BytesParser().parsebytes(email_msg)
		return cls.serialize_from_email_message(parsed_msg)

	def to_dict(self):
		msg = {
			"date": self._datetime,
			"message_id": self._message_id,
			"from": dataclasses.asdict(self._from),
			"to": [ dataclasses.asdict(addr) for addr in self._to ],
		}
		if self.subject is not None:
			msg["subject"] = self.subject
		if len(self._cc) > 0:
			msg["cc"] = [ dataclasses.asdict(addr) for addr in self._cc ]
		if len(self._bcc) > 0:
			msg["bcc"] = [ dataclasses.asdict(addr) for addr in self._bcc ]
		if self._text is not None:
			msg["text"] = self._text
		if self._html is not None:
			msg["html"] = self._html
		return msg

	def __format__(self, fmt_str: str):
		if self.subject is None:
			text = "-no subject-"
		else:
			text = f"\"{self.subject}\""
		text += f" from {self._from.mail}"
		text += f" to {', '.join(mail.mail for mail in self._to)}"
		return text



#	mailcoil - Effortless, featureful SMTP
#	Copyright (C) 2011-2025 Johannes Bauer
#
#	This file is part of mailcoil.
#
#	mailcoil is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	mailcoil is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with mailcoil; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import io
import enum
import subprocess
import email.parser

class CMS():
	class HashFunction(enum.Enum):
		sha1 = "sha1"
		sha256 = "sha256"
		sha384 = "sha384"
		sha512 = "sha512"
		sha3_256 = "sha3-256"
		sha3_384 = "sha3-384"
		sha3_512 = "sha3-512"

	class Cipher(enum.Enum):
		AES128 = "aes128"
		AES256 = "aes256"

	def __init__(self, openssl_binary: str = "openssl", hashfnc: HashFunction = HashFunction.sha256, cipher: Cipher = Cipher.AES256):
		self._openssl_binary = openssl_binary
		self._hashfnc = hashfnc
		self._cipher = cipher
		self._signer_certfile = None
		self._signer_keyfile = None
		self._signer_keyform = None
		self._detach_signature = True
		self._ca_certfile = None
		self._encrypt_recipient_certfiles = [ ]

	def sign(self, signer_certfile: str, signer_keyfile: str, ca_certfile: str | None = None, keyform: str = "pem", detach_signature: bool = True):
		self._signer_certfile = signer_certfile
		self._signer_keyfile = signer_keyfile
		self._ca_certfile = ca_certfile
		self._signer_keyform = keyform
		self._detach_signature = detach_signature
		return self

	def encrypt(self, *recipient_certfiles: tuple[str]):
		self._encrypt_recipient_certfiles += recipient_certfiles
		return self

	def _encrypt(self, message: bytes):
		cmd = [ self._openssl_binary, "cms", "-encrypt" ]
		cmd += [ f"-{self._cipher.value}" ]
		cmd += self._encrypt_recipient_certfiles
		return subprocess.check_output(cmd, input = message)

	def _sign(self, message: bytes):
		cmd = [ self._openssl_binary, "cms", "-sign" ]
		cmd += [ "-signer", self._signer_certfile ]
		if not self._detach_signature:
			cmd += [ "-nodetach" ]
		cmd += [ "-inkey", self._signer_keyfile, "-keyform", self._signer_keyform ]
		cmd += [ "-md", self._hashfnc.value ]
		if self._ca_certfile is not None:
			cmd += [ "-certfile", self._ca_certfile ]
		return subprocess.check_output(cmd, input = message)

	def process(self, msg: "MIMEBase"):
		# Order of operations in S/MIME is AtE.
		if self._signer_keyfile is not None:
			# Sign
			msg = self._sign(bytes(msg))
		if len(self._encrypt_recipient_certfiles) > 0:
			# Encrypt
			msg = self._encrypt(bytes(msg))

		# Then parse
		f = io.BytesIO(msg)
		msg = email.parser.BytesParser().parse(f)
		return msg
